/*
 * $LastChangedBy$
 * $LastChangedDate$
 * $LastChangedRevision$
 * $HeadURL$
 *
 * Copyright 2014 BlueConic Inc./BlueConic B.V. All rights reserved.
 */
import Foundation
import UIKit
private func < <T: Comparable>(lhs: T?, rhs: T?) -> Bool {
    switch (lhs, rhs) {
    case let (l?, r?):
        return l < r
    case (nil, _?):
        return true
    default:
        return false
    }
}

private func > <T: Comparable>(lhs: T?, rhs: T?) -> Bool {
    switch (lhs, rhs) {
    case let (l?, r?):
        return l > r
    default:
        return rhs < lhs
    }
}

public class ListenerUtil {
    /**
     Checks if the collected content contains a specified word. Also checks how it should match according to blueconic configuration.
     - parameter rule: The listener rule
     - parameter content: The content from the UI element
     - parameter interests: The interests keywords, which could be matching with the content
     
     - returns: true if content containing a matching interest value
     */
    public class func contentContainsWord(_ rule: JSON, content: [String], interests: [String]) -> Bool {
        var matchingtype = rule[PluginConstants.Listener.TAG_CONTAINS_MATCHES].string
        if matchingtype == nil {
            matchingtype = PluginConstants.Listener.CONTAINS
        }
        
        if matchingtype == PluginConstants.Listener.EMPTY {
            return content.count == 0
        }
        
        if content.count < 1 {
            return false
        }
        
        // join the content and lowercase it
        let contentString: String = content.joined(separator: " ").trimmingCharacters(in: .whitespaces).lowercased()
        
        for untrimmedInterest in interests {
            let interest = untrimmedInterest.trimmingCharacters(in: .whitespaces).lowercased()
            if interest == PluginConstants.Listener.PRE_ANY ||  interest == PluginConstants.Listener.PRE_HEADER {
                return true
            } else if rule[PluginConstants.Listener.TAG_RULE_TYPE].stringValue == PluginConstants.Listener.RULETYPE_SCORE_URL ||
                rule[PluginConstants.Listener.TAG_RULE_TYPE].stringValue == PluginConstants.Listener.RULETYPE_INTEREST_URL {
                if matchingtype == PluginConstants.Listener.CONTAINS {
                    // In swift, we shouldn't need to escape any characters within the string
                    if contentString.range(of: interest) != nil {
                        return true
                    }
                } else if contentString == interest {
                    return true
                }
            } else if matchingtype == PluginConstants.Listener.CONTAINS && contentString.range(of: interest) != nil {
                return true
            } else if matchingtype == PluginConstants.Listener.MATCHES {
                for c in content {
                    if c.trimmingCharacters(in: .whitespaces).lowercased() == interest {
                        return true
                    }
                }
            }
        }
        
        return false
    }
    
    public class func getWords(_ rule: JSON) -> [String] {
        return self.getWords(rule[PluginConstants.Listener.TAG_WORDS].arrayObject as? [String])
    }
    
    /**
     Returns words if available. If not, returns "pre_any" denoting any interest can be matched
     - parameter words:  an optional string array
     - returns:  a string array containing the provided words or one word "pre_any"
     */
    public class func getWords(_ words: [String]?) -> [String] {
        if words != nil && words?.count > 0 {
            return words!
        }
        return [PluginConstants.Listener.PRE_ANY]
    }
    
    /**
     Returns content matching the selector. It iterates over subcontent to collect all content strings.
     - parameter selector: a string containing a selector, this should be in the form of "#<selector>", for example "#exampleButton". Only simple id matches are supported.
     - returns: a string array containing all found content
     */
    public class func getContent(_ selector: String, context: [String]?) -> [String] {
        if selector == PluginConstants.Listener.PRE_CONTEXT {
            return context ?? []
        }

        var result: Set<String> = []
        if selector.range(of: "jQuery(") != nil {
            // we don't want to do anything with jquery
            return []
        } else {
            let contentElement: UIView?
            if selector == PluginConstants.Listener.ANY ||  selector == PluginConstants.Listener.PRE_HEADER {
                contentElement = BlueConicClient.getInstance(nil).getViewController()?.view
            } else {
                // we first get the UIView element
                contentElement = BlueConicClient.getInstance(nil).getView(selector)
            }
            
            
            if contentElement != nil {
                // aggregate all content wihtin a uiview using a closure, applyToSubViews recursively navigates all subviews and executes the provided closure
                result = applyToSubViews(contentElement, apply: {(el: UIView) -> String? in
                    if el.responds(to: Selector(PluginConstants.Listener.TEXT_KEY)) {
                        return el.value(forKey: PluginConstants.Listener.TEXT_KEY) as? String
                    } else {
                        return nil
                    }
                })
            }
        }
        
        return (Array(result)).filter({$0.count > 0})
    }
    
    /**
     This function recursively iterates over subviews depth first and applies a provided closure returning set containing all results. Since we return a set, no duplicates are returned.
     
     - parameter element: a UIView element to inspect, if it has subviews these are inspected as well
     - parameter apply: a closure to apply to each uiview found
     
     - returns: A set of strings representing what is aggregated over the subviews.
     */
    public class func applyToSubViews(_ element: UIView?, apply: (UIView) -> (String?)) -> Set<String> {
        // recursively iterate over subviews
        var resultList: Set<String> = []
        
        if let element = element {
            if let result = apply(element) {
                resultList.insert(result)
            }
            let subViews: [UIView] = element.subviews
            for subView in subViews {
                resultList = resultList.union(applyToSubViews(subView, apply: apply))
            }
        }
        return resultList
    }
    
    /**
     Retrieves a selector from rule json
     - parameter contentArea: A piece of json containing a selector
     - returns:  A selector, if nothing is found returns nil
     */
    public class func getSelector(_ contentArea: JSON?) -> String? {
        var selector: String?
        if let contentArea = contentArea {
            if let selectorString = contentArea["selector"].string {
                if selectorString == PluginConstants.Listener.PRE_ANY || selectorString == PluginConstants.Listener.PRE_BODY || selectorString == PluginConstants.Listener.PRE_MOBILE {
                    selector = PluginConstants.Listener.ANY
                } else {
                    selector = selectorString
                }
            }
        }
        return selector
    }
    
    public class func JSONStringify(_ value: Any, prettyPrinted: Bool = false) -> String {
        let options = prettyPrinted ? JSONSerialization.WritingOptions.prettyPrinted : JSONSerialization.WritingOptions(rawValue: 0)
        if JSONSerialization.isValidJSONObject(value) {
            
            if let data = try? JSONSerialization.data(withJSONObject: value, options: options),
                let string = NSString(data: data, encoding: String.Encoding.utf8.rawValue) {
                return string as String
            }
            
        } else {
            BCLog("Not a valid JSON object", flag: Constants.Flag.WARN)
        }
        return ""
    }
    
    /**
     Helper function to get the profile properties from the object
     
     - parameter parameters:  A Dictionary mapping strings to string arrays.
     
     - returns: an optional string array containing one or more values
     */
    public class func getProfilePropertyFromParameters(_ parameters: [String: [String]]) -> String? {
        if let propertyArray = parameters["property"],
            let jsonArray = ListenerUtil.stringToJson(propertyArray[0]) as? [[String: String]],
            let property = jsonArray[0]["profileproperty"], !property.isEmpty {
            return property
        }
        return nil
    }
    
    /**
     Returns a dictionary from the parameters
     
     - parameter context: The interaction context
     - parameter key: the parameter key containing the dictionary
     
     - returns: a Dictionary mapping string to Any array
     */
    public class func getDictionaryFromParameters(_ context: InteractionContext?, key: String) -> [String: [Any]]? {
        if let parameters: [String: [String]] = context?.getParameters(),
            let values = parameters[key], values.count > 0 {
            return ListenerUtil.stringToJson(values[0]) as? [String: [Any]]
        }
        return nil
    }
    
    /**
     Returns a dictionary from the parameters
     
     - parameter context: The interaction context
     - parameter key: the parameter key containing the dictionary
     
     - returns: a Dictionary mapping string to Any array
     */
    public class func getArrayFromParameters(_ context: InteractionContext?, key: String) -> [[String: Any]]? {
        if let values = context?.getValueFromParameters(key: key),
            let items = ListenerUtil.stringToJson(values) as? [[String: Any]] {
            return items
        }
        return nil
    }
    
    /**
     Get a value from the parameters. Retrieves the parameters from the blueconic client object itself.
     - parameter context: The interaction context
     - parameter key: the parameter key to retrieve values from the parameters
     - returns: a parameter value
     */
    public class func getValueFromParameters(_ context: InteractionContext?, key: String) -> String {
        if let parameters: [String: [String]] = context?.getParameters(),
            let values = parameters[key], values.count > 0 {
            return values[0]
        }
        
        return ""
    }
    
    
    /**
     Get a value from the parameters. Retrieves the parameters from the blueconic client object itself.
     - parameter context: The interaction context
     - parameter key: the parameter key to retrieve values from the parameters
     - returns: a parameter value
     */
    public class func getValueFromDictionary(_ dictionary: [String: [String]], key: String) -> String {
        if let values = dictionary[key], values.count > 0 {
            return values[0]
        }
        
        return ""
    }
    
    public class func stringToJson(_ json: String) -> Any? {
        do {
            return try JSONSerialization.jsonObject(with: json.data(using: String.Encoding.utf8)!, options: JSONSerialization.ReadingOptions(rawValue: 0))
        } catch let error as NSError {
            
            BCLog("JSON error occurred: \(error.localizedDescription)", flag: Constants.Flag.ERROR)
        }
        
        return nil
    }


    /**
     Return current date/time in milisecconds
     - returns: the current date/time in milliseconds
     */
    public class func getCurrentTime() -> String {
        return Int64(Date().timeIntervalSince1970 * 1000).description
    }


    /**
     Return current date in milliseconds
     - returns: the current date (without any time 00:00:00.000) in milliseconds
     */
    public class func getCurrentDate() -> String {
        let timeZone = TimeZone.current
        let timeIntervalWithTimeZone = Date.timeIntervalSinceReferenceDate + Double(timeZone.secondsFromGMT())
        let timeInterval = floor(timeIntervalWithTimeZone / 86400) * 86400

        return Int64(Date(timeIntervalSinceReferenceDate: timeInterval).timeIntervalSince1970 * 1000).description
    }
    
    private class func BCLog(_ message: String, flag: String = Constants.Flag.INFO) {
        BlueConicConfiguration.log(message, tag: Constants.Debug.LISTENER_UTIL, flag: flag)
    }
}
