/*
* $LastChangedBy$
* $LastChangedDate$
* $LastChangedRevision$
* $HeadURL$
*
* Copyright 2014 BlueConic Inc./BlueConic B.V. All rights reserved.
*/

import Foundation

class BlueConicCommitLog: Codable {
    private var _properties: [BlueConicPropertyCommitEntry]
    private var _events: [BlueConicEventCommitEntry]

    private var _modified: Bool = false

    /**
    Default init() for BlueConicCommitLog
    */
    init() {
        self._properties = [BlueConicPropertyCommitEntry]()
        self._events = [BlueConicEventCommitEntry]()

        self._modified = true
    }

    /**
     Initialize with entries
     - parameter entries: The entries that are committed to log
     */
    init(properties: [BlueConicPropertyCommitEntry], events: [BlueConicEventCommitEntry]) {
        self._properties = properties
        self._events = events
        self._modified = false
    }

    /**
    Sets the values for the property with the specified name.

    - parameter name: The property name
    - parameter values: The String array containing the new values
    */
    func setProperties(_ name: String, values: [String]) {
        if name == "" || values.count == 0 {
            return
        }

        // Check if the PropertyCommitEntry already exists, if so override the values
        if let exitingEntry = getPropertyCommitEntry(name) {
            exitingEntry.setValues(values)
        } else {
            // Create a new set property entry with the specified values
            let entry = BlueConicPropertyCommitEntry(propertyIdentifier: name, type: OperationType.SET, values: values)
            self._properties.append(entry)
        }

        // Store a new entry and overwrite any existing entry
        self.setLastUpdate()
    }

    /**
    Adds values to the property with the specified name.

    - parameter name: The property name
    - parameter values: The String array containing the additional values
    */
    func addProperties(_ name: String, values: [String]) {
        if name == "" || values.count == 0 {
            return
        }

        // Check if the PropertyCommitEntry already exists, if so then ADD the value to the entry
        if let existingEntry = getPropertyCommitEntry(name) {
            let existingValues: [String] = existingEntry.getValues()
            for value in values {
                if !existingValues.contains(value) {
                    existingEntry.addValue(value)
                }
            }
        } else {
            // Create a new add PropertyCommitEntry with the specified values
            let entry = BlueConicPropertyCommitEntry(propertyIdentifier: name, type: OperationType.ADD, values: values)
            self._properties.append(entry)
        }
        self.setLastUpdate()
    }

    /**
     Add a created even to the entries, based on event-type and interactionId

     - parameter eventType: The eventype, one of {VIEW, CLICK, CONVERSION}
     - parameter interactionId: The position identifier of an element in the app
    */
    func createEvent(_ eventType: EventType, interactionId: String) {
        if interactionId == "" {
            return
        }
        // Update the event.
        self.updateEvent(eventType, interactionId: interactionId, amount: 1)

        self.setLastUpdate()
    }
    /**
     Update event, based on event-type and interactionId.
     Checks if the event already exists, then it will increase the count.
     else it will add a new event commit entry.

     - parameter eventType: The eventype, one of {VIEW, CLICK, CONVERSION}.
     - parameter interactionId: The position identifier of an element in the app.
     - parameter amount: The amount that the count has to be increased.
    */
    private func updateEvent(_ eventType: EventType, interactionId: String, amount: Int) {
        // Check if the EventCommitEntry already exists if so then increase the count of the entry
        if let existingEntry = getEventCommitEntry(eventType, interactionId: interactionId) {
            existingEntry.increaseCount(amount)
        } else {
            // Create a new event entry and set the count to 1
            let newEntry = BlueConicEventCommitEntry(eventType: eventType, interactionId: interactionId)
            if amount > 1 {
                newEntry.increaseCount(amount-1)
            }
            self._events.append(newEntry)
        }
    }

    /**
     Returns the dictionary of operations in the commitlog.

     - parameter type: The of the operation, add or set
     - returns: The dicionary of matching entries, being a map of Strings to arrays of Strings.
    */
    func getProperties() -> [BlueConicPropertyCommitEntry] {
        return self._properties
    }
    func getProperties(_ type: OperationType) -> [String: BlueConicPropertyCommitEntry] {
        var result = [String: BlueConicPropertyCommitEntry]()
        for entry in self._properties {
            if entry.getType() == type.rawValue {
                result[entry.getId()] = entry
            }
        }
        return result
    }

    /**
    Returns all Created events in the commitlog.

    - returns: list with all BlueConicEventCommitEntries
    */
     func getEvents() -> [BlueConicEventCommitEntry] {
        return self._events
    }
    /**
    Returns all the entries in the commitlog.

    - returns: list with all BlueConicEventCommitEntries
    */
     func getCount() -> Int {
        return self._events.count + self._properties.count
    }

    /**
    Returns a BlueConicPropertyCommitEntry when there is already an entry in the commmitlog, if not then return nil.

    - parameter propertyId:: identifier of a profile property
    - returns: BlueConicPropertyCommitEntry or nil
    */
     func getPropertyCommitEntry(_ propertyId: String) -> BlueConicPropertyCommitEntry? {
        if propertyId == "" {
            return nil
        }

        // Check if an entry is a PropertyCommitEntry and if the id is equal to propertyId
        return self._properties.first(where: {$0.getId() == propertyId})
    }

    /**
    Returns a BlueConicEventCommitEntry when there is already an entry in the commmitlog, if not the return nil.

    - parameter eventType:: could be {VIEW, CLICK, CONVERSION}
    - parameter interactionId:: position identifier of the element in the app
    - returns: BlueConicEventCommitEntry or nil
    */
    func getEventCommitEntry(_ eventType: EventType, interactionId: String) -> BlueConicEventCommitEntry? {
        if interactionId == "" {
            return nil
        }

        // Find the first event that matches the interactino id and event type
        return self._events.first(where: {$0.getId() == interactionId && $0.getType() == eventType.rawValue})
    }

    /**
    Clears all entries in the commit log.
    */
    func clearAll() {
        self._properties.removeAll()
        self._events.removeAll()
    }

    /**
     Merge the commitlogs, the parameter is the newest commitlog,
     so setters will be override the old set values.
     - parameter commitLog: The current commit log.
    */
    func mergeCommitLog(_ commitLog: BlueConicCommitLog) {
        // Return if there are no entries to merge
        if commitLog.getCount() == 0 {
            return
        }

        // Merge the commitlog entries if both lists contain entries.
        for entry in commitLog.getEvents() {
            self.updateEvent(EventType(rawValue: entry.getType())!, interactionId: entry.getId(), amount: entry.getCount())
        }
        for entry in commitLog.getProperties() {
            if entry.getType() == OperationType.SET.rawValue {
                self.setProperties(entry.getId(), values: entry.getValues())
            } else if entry.getType() == OperationType.ADD.rawValue {
                self.addProperties(entry.getId(), values: entry.getValues())
            }
        }

        commitLog.clearAll()
    }

    /**
    Reads the modified flag and resets it.

    - returns: The value of the modified flag before it was reset.
    */
    func checkModified() -> Bool {
        let result: Bool = _modified
        self._modified =  false
        return result
    }

    /**
    Set the modified value to true, so the updater knows the commit log has changed
    */
    func setLastUpdate() {
        self._modified = true
    }
}

//
//  BlueConicCommitEntry
//
class BlueConicCommitEntry: Codable {
    private let _identifier: String

    /**
     Initializer with identifier
     - parameter id the identifier of the commit entry
     */
     init(_ id: String) {
        self._identifier = id
    }

    // Entry implements
     func getType() -> String {
        return ""
    }
     func getId() -> String {
        return self._identifier
    }
}

//
//  BlueConicPropertyCommitEntry
//
class BlueConicPropertyCommitEntry: BlueConicCommitEntry {
    private let _type: OperationType
    private var _values: [String]

    enum PropertyEntryKeys: String, CodingKey {
        case _identifier
        case _type
        case _values
    }

    /**
    Constructor.

    - parameter propertyIdentifier: The property identifier.
    - parameter type: The type of operation.
    - parameter values: The values.
    */
    init(propertyIdentifier: String, type: OperationType, values: [String]) {
        self._type = type
        self._values = values
        super.init(propertyIdentifier)
    }

    /**
    Implementation of Codable.
    init() with a Coder set on the entries for BlueConicPropertyCommitEntry.
     - parameter decoder: The decoder for the commit log entries.
    */
    convenience required init(from decoder: Decoder) throws {
        let container = try decoder.container(keyedBy: PropertyEntryKeys.self)

        let identifier = try container.decode(String.self, forKey: ._identifier)
        let type = try container.decode(String.self, forKey: ._type)
        let values = try container.decode([String].self, forKey: ._values)
        self.init(propertyIdentifier: identifier, type: OperationType(rawValue: type)!, values: values)
    }

    /**
     Implementation of Codable.
     encode with Coder to get the entries from BlueConicCommitLog.
     - parameter coder: THe coder for the commit log entries.

    */
    override func encode(to encoder: Encoder) throws {
        var container = encoder.container(keyedBy: PropertyEntryKeys.self)
        try container.encode(self.getId(), forKey: ._identifier)
        try container.encode(self._values, forKey: ._values)
        try container.encode(self.getType(), forKey: ._type)
    }

    /**
    Returns the type of operation.

    - returns: The type of operation. This can be {'add', 'set'}.
    */
    override func getType() -> String {
        return self._type.rawValue
    }

    /**
    Returns the values.

    - returns: The values.
    */
     func getValues() -> [String] {
        return self._values
    }

    /**
    Set the values

    - parameter values: An array of property values
    */
     func setValues(_ values: [String]) {
        if values.count > 0 {
            self._values = values
        }
    }

    /**
    Add a value

    - parameter value: A property value
    */
     func addValue(_ value: String) {
        self._values.append(value)
    }
}

//
//  BlueConicEventCommitEntry
//
class BlueConicEventCommitEntry: BlueConicCommitEntry {
    private let _type: EventType
    private var _count: Int = 1

    /**
    Constructor.

    - parameter eventType:     A EventType enum {VIEW, CLICK, CONVERSION}.
    - parameter interactionId: The Id of an interacion.
    */
    init(eventType: EventType, interactionId: String) {
        self._type = eventType
        super.init(interactionId)
    }


    enum EventEntryKeys: String, CodingKey {
        case _identifier
        case _type
        case _count
    }

    /**
     Implementation of Codable.
     init() with a Coder set on the entries for BlueConicEventCommitEntry.
     - parameter decoder: for the commit log entries.
     */
    convenience required init(from decoder: Decoder) throws {
        let container = try decoder.container(keyedBy: EventEntryKeys.self)

        let identifier = try container.decode(String.self, forKey: ._identifier)
        let type = try container.decode(String.self, forKey: ._type)
        let count = try container.decode(Int.self, forKey: ._count)

        self.init(eventType: EventType(rawValue: type)!, interactionId: identifier)
        self._count = count
    }

    /**
     Implementation of Codable.
     encode with Coder to get the entries from BlueConicEventCommitEntry.
     - parameter coder: for the commit log entries.
     */
    override func encode(to encoder: Encoder) throws {
        var container = encoder.container(keyedBy: EventEntryKeys.self)
        try container.encode(self.getId(), forKey: ._identifier)
        try container.encode(self.getType(), forKey: ._type)
        try container.encode(self.getCount(), forKey: ._count)
    }

    /**
    Returns the type of the event

    - returns: The type of the event, This can be { "VIEW", "CLICK",  "CONVERSION" }
    */
    override func getType() -> String {
        return self._type.rawValue
    }

    /**
    Returns the current count of the event

    - returns: A count how many times this event is called.
    */
     func getCount() -> Int {
        return self._count
    }

    /**
    Increasing the count
    - parameter amount:  The amount of
    */
     func increaseCount(_ amount: Int) {
        self._count += amount
    }
}
