/*
* $LastChangedBy$
* $LastChangedDate$
* $LastChangedRevision$
* $HeadURL$
*
* Copyright 2014 BlueConic Inc./BlueConic B.V. All rights reserved.
*/

import Foundation
import UIKit

public enum EventType: String {
    case CLICK
    case VIEW
    case CONVERSION
}

public enum OperationType: String {
    case ADD
    case SET
}

public enum CallId: Int {
    case profile, addProperties, setProperties, getProperties, interactions, events
}

class BlueConicConfiguration {

    private var _mobileSessionId: String?
    private var _userName: String?

    private var _overruleHostname: String?

    init() {
        self.loadSimulatorData()
    }

    /**
    Method returns the Hostname from mainbundle infoDictionary

    - returns: the hostname or nil
    */
    func getHostName() -> String {

        if let overrule = self._overruleHostname {
            return overrule
        }

        var result = ""
        if let hostName = Bundle.main.infoDictionary?[Constants.HOST_NAME] as? String {
            result = hostName
            if hostName.hasSuffix("/") {
                result = String(hostName.dropLast())
            }
        }

        return result
    }

    /**
    Setter for the hostname.

    - parameter hostName: e.g. 'https://example.blueconic.com'
    */
    func setHostName(_ hostName: String) {
        self._overruleHostname = hostName
    }

    /**
    Get the PackageName of the Mobile app which is using BlueConicClient
    PackageName is something like com.blueconictest.BCTestApp

    - returns:   a string with UTF8 encoded String (e.q. com.blueconic.TestBCTestApp)
    */
    class func getPackageName() -> String {
        if let info = Bundle.main.infoDictionary,
            let appId = info[Constants.MobileApp.BUNDLE_IDENTIFIER] as? String {
                let encodeAppId: String? = appId.addingPercentEncoding(withAllowedCharacters: CharacterSet.urlFragmentAllowed)!
            if let result = encodeAppId {
                return result
            }
        }
        return ""
    }

    /**
    Method returns an URL for which the BCSessionId cookie needs to be set(or retrieved).

    - parameter hostName: The hostname.
    - parameter domainGroup: The domain group.
    - returns: The URL.
    */
    class func getDomaingGroupUrl(_ hostName: String, domainGroup: String?, zoneId: String?) -> String {
        let domain: String = domainGroup != nil ? domainGroup! : Constants.Connector.DOMAIN_VALUE
        let zone: String = zoneId != nil ? zoneId! + "/" : ""
        return hostName + "/DG/" + domain + "/rest/rpc/json/" + zone + "?"
    }

    /**
    Method that returns the Debug Mode, which is set in the main bundle

    - returns:   the debug mode, when true it will log the debug information to the console
    */
    class func getDebugMode() -> Bool {

        if let result: Bool = Bundle.main.infoDictionary?[Constants.DEBUG_MODE] as? Bool {
            return result
        }
        return false
    }
    
    /**
    Method that returns whether session and content storage are shared, which is set in the main bundle

    - returns:   whether session and content storage are shared
    */
    class func getShared() -> Bool {

        if let result: Bool = Bundle.main.infoDictionary?[Constants.SHARED] as? Bool {
            return result
        }
        return true
    }

    /**
     Log a debug message
     - parameter: message The mesage of the log
     - parameter: tag Define a message with a tag, by default we use BC_CLIENT
     - parameter: flag Define a message with a flag, by default we use INFO
     */
    class func log(_ message: String, tag: String = Constants.Debug.CLIENT, flag: String = Constants.Flag.INFO) {
        if getDebugMode() {
            print("[\(printDate())] \(flag) [\(tag)]: \(message)")
        }
    }

    /**
     Returns a printable date in format "yyyy-MM-dd HH:mm:ss"
     - returns:   a printable date
     */
    private class func printDate() -> String {
        let formatter = DateFormatter()
        formatter.dateFormat = Constants.DATE_FORMAT
        return formatter.string(from: Date())

    }

    /*
    Set Simulator Data
    When a QR code is scanned it will call this function to save the username and mobile session id
    During this session it will add the username and mobile session id to each method call

    :param: userName this value will be the user's e-mailaddress who is logged at the BlueConic Simulator
    :param: mobileSessionId is an unique id, which allows requests to be send as "Simulated"
    */
    func setSimulatorData(_ userName: String, mobileSessionId: String) {
        self._userName = userName
        self._mobileSessionId = mobileSessionId
        self.saveSimulatorData(userName, mobileSessionId: mobileSessionId)
    }

    /**
     Write simulator data to a specified file
     - parameter profileId: The profile identifier retrieved from the server
     */
    func saveSimulatorData(_ userName: String, mobileSessionId: String) {
        BlueConicDataManager.writeStringToFile(data: userName, fileName: Constants.Files.SIMULATOR_USERNAME)
        BlueConicDataManager.writeStringToFile(data: mobileSessionId, fileName: Constants.Files.SIMULATOR_MOBILESESSION)
    }


    /**
     Load existing simualtor data from a specified file
     */
    func loadSimulatorData() {
        self._userName = BlueConicDataManager.loadStringFromFile(fileName: Constants.Files.SIMULATOR_USERNAME)
        self._mobileSessionId = BlueConicDataManager.loadStringFromFile(fileName: Constants.Files.SIMULATOR_MOBILESESSION)
    }

    /**
     Returns the username from the connected simulator
     */
    func getSimulatorUsername() -> String? {
        return self._userName
    }

    /**
     Returns the session id from the connected simulator
     */
    func getSimulatorSessionId() -> String? {
        return self._mobileSessionId
    }

    /**
     Whether the device is connected to a simulator session
     returns true when it has connected to the simulator
     */
    func isConnectedToSimulator() -> Bool {
        return Bool(self._userName != nil && self._mobileSessionId != nil)
    }

    // MARK: Visable ViewController

    /**
    Method that gets the current visible ViewController
    It recieves the root ViewController from sharedApplication
    and loops through it until the current active View Controller is found

    - returns:   the active view controller from the Application.
    */
    class func getVisibleViewController() -> UIViewController? {
        if let rootViewController: UIViewController = UIApplication.shared.keyWindow?.rootViewController {
            return getVisibleViewControllerFrom(rootViewController)
        }
        return nil
    }

    class func getVisibleViewControllerFrom(_ rootViewController: UIViewController) -> UIViewController {
        if let nav = rootViewController as? UINavigationController {
            return getVisibleViewControllerFrom(nav.visibleViewController!)
        }
        if let tab = rootViewController as? UITabBarController {
            if let selected = tab.selectedViewController {
                return getVisibleViewControllerFrom(selected)
            }
        }
        if let presented = rootViewController.presentedViewController {
            return getVisibleViewControllerFrom(presented)
        }
        return rootViewController
    }

    class func parametersToDictionary(_ parameters: [[String: Any]]) -> [String: [String]] {
        var result = [String: [String]]()
        if parameters.count > 0 {
            for parameterDictionary in parameters {
                var paramKey = ""
                var paramValue = [String]()
                for (key, value) in parameterDictionary {
                    if key == "id" {
                        paramKey = value as? String ?? ""
                    } else if key == "value" {
                        paramValue = value as? [String] ?? [String]()
                    }
                }
                result[paramKey] = paramValue
            }
        }
        return result
    }
}
