/*
* $LastChangedBy$
* $LastChangedDate$
* $LastChangedRevision$
* $HeadURL$
*
* Copyright 2014 BlueConic Inc./BlueConic B.V. All rights reserved.
*/

import Foundation

// METHOD_TYPE used for HTTPMethod
enum METHOD_TYPE: String {
    case GET
    case POST
    case PUT
    case DELETE
}

class BlueConicConnector {
    private var _mobileSessionId: String?
    private var _userName: String?
    private var _session: URLSession

    private var _originalRequest: URLRequest?
    private var _data: Data!

    private var _error: Error?
    private var _response: URLResponse?
    
    private enum BlueConicError: Error {
        case MaxRetriesReached
    }

    // Retry properties
    private var _retryCount: Int = 0
    private let MAX_RETRIES = 3

    init(session: URLSession, userName: String?, mobileSessionId: String?) {
        self._userName = userName
        self._mobileSessionId = mobileSessionId
        self._session = session
        self._retryCount = 0
    }

    /**
     Invokes a number of methods.

     - parameter baseUrl: The base url of the http request.
     - parameter calls: The array of method call objects.
     - parameter screenName: The screen name of the current view controller.
     - returns: The mapping of method call ids to the correspondig result dictionary or nil of the invocation failed.
    */
    func execute(_ baseUrl: String, calls: [BlueConicRequestCommand], screenName: String?) -> (response: [String: Any]?, error: Error?) {
        // Make sure cookies are accepted
        let cookieStore: HTTPCookieStorage = HTTPCookieStorage.shared
        if cookieStore.cookieAcceptPolicy == HTTPCookie.AcceptPolicy.never {
            // Set the strategy to the minimum required value
            cookieStore.cookieAcceptPolicy = HTTPCookie.AcceptPolicy.onlyFromMainDocumentDomain
        }
        self._data = Data()

        // Create the request that should be executed
        //let request: URLRequest = self.createRequest(baseUrl, commands: calls, screenName: screenName)
        var requestUrl = baseUrl
        if let lastChar = baseUrl.last, lastChar == "?" {
            requestUrl += getPostParameters()
        }

        BCLog("Request URL to Blueconic: \(requestUrl)")
        
        guard let request: URLRequest = self.createRequest(requestUrl, commands: calls, screenName: screenName) else {
            return (nil, nil)
        }
        
        self._originalRequest = request
        
        // Execute the request synchronously
        let semaphore = DispatchSemaphore(value: 0)
        let task = self._session.dataTask(with: request) { (data, response, error) in
            self._response = response
            self._data = data
            self._error = error
            semaphore.signal()
        }

        task.resume()
        // Set a timeout of 10 seconds
        let timeout = DispatchTime.now() + .seconds(10)
        _ = semaphore.wait(timeout: timeout)

        // Check if the response is a redirect with the status code 300
        if let httpResponse: HTTPURLResponse = self._response as? HTTPURLResponse,
            httpResponse.statusCode == 300,
            let data = self._data,
            let result = parseResponse(data).result,
            let location = result["location"] as? String {
                return executeRedirect(location, calls: calls, screenName: screenName)
        }

        // Check if the response is a redirect with another URL
        if let responseUrl: String = self._response?.url?.absoluteString, responseUrl != requestUrl {
            return executeRedirect(responseUrl, calls: calls, screenName: screenName)
        }

        // Retry on failure
        if let error = self._error, error._code == -1005 {
            return executeFailure(baseUrl, calls: calls, screenName: screenName)
        }

        // Check weather the call completed successfully
        if self._error != nil {
            return(nil, self._error)
        } else if self._response == nil || !(self._response?.isKind(of: HTTPURLResponse.self) != nil) {
            return(nil, nil)
        }
        
        let httpResponse = self._response as? HTTPURLResponse

        // Check the response code
        if httpResponse?.statusCode != 200 {
            return(nil, nil)
        }
        
        if !BlueConicConfiguration.getShared() {
            let cookieStorage = self._session.configuration.httpCookieStorage as! BlueConicCookieStorage
            
            
            if let url = httpResponse?.url, let allHeaderFields = httpResponse?.allHeaderFields as? [String : String] {
                let cookies = HTTPCookie.cookies(withResponseHeaderFields: allHeaderFields, for: url)
                cookieStorage.addCookies(cookies)
            }
        }
        
        if let data = self._data {
            //Parse the response into a dictionary
            let response = parseResponse(data)

            if response.error != nil || response.result == nil {
                return(nil, response.error)
            }

            // Return the results for each call in a dictionary
            return (response.result, nil)
        }
               
        BCLog("Request with an empty response")
        return (nil, nil)
    }

    /**
     Retry executing the number of methods, when a redirect from the server has taken place.

     - parameter baseUrl: The base url of the http request.
     - parameter calls: The array of method call objects.
     - parameter screenName: The screen name of the current view controller.
     - returns: The mapping of method call ids to the correspondig result dictionary or nil of the invocation failed.
     */
    func executeRedirect(_ baseUrl: String, calls: [BlueConicRequestCommand], screenName: String?) -> (response: [String: Any]?, error: Error?) {
        BCLog("Redirect to URL: \(baseUrl)")

        return redirect(baseUrl, calls: calls, screenName: screenName)
    }

    /**
     Retry executing the number of methods, when an error has taken place.

     - parameter baseUrl: The base url of the http request.
     - parameter calls: The array of method call objects.
     - parameter screenName: The screen name of the current view controller.
     - returns: The mapping of method call ids to the correspondig result dictionary or nil of the invocation failed.
     */
    func executeFailure(_ baseUrl: String, calls: [BlueConicRequestCommand], screenName: String?) -> (response: [String: Any]?, error: Error?) {
        BCLog("Sending data to BlueConic. Retry attempt: \(self._retryCount)", flag: Constants.Flag.WARN)

        self._error = nil
        return redirect(baseUrl, calls: calls, screenName: screenName)
    }

    /**
     Retry executing the number of methods

     - parameter baseUrl: The base url of the http request.
     - parameter calls: The array of method call objects.
     - parameter screenName: The screen name of the current view controller.
     - returns: The mapping of method call ids to the correspondig result dictionary or nil of the invocation failed.
     */
    func redirect(_ baseUrl: String, calls: [BlueConicRequestCommand], screenName: String?) -> (response: [String: Any]?, error: Error?) {
        if self._retryCount > MAX_RETRIES {
            return (nil, BlueConicError.MaxRetriesReached)
        }
        self._retryCount += 1
        return execute(baseUrl, calls: calls, screenName: screenName)
    }

    /**
     Parses the response data of an rpc resource call.

    - parameter receiveData: The response bytes.
    - returns: The mapping of method call ids to the correspondig result dictionary or nil of the invocation failed.

    */
    func parseResponse(_ receiveData: Data) -> (result: [String: Any]?, error: Error?) {
        // Get the response body as String
        var result: [String: Any] = [:]
        var response: String = String(data: receiveData, encoding: .utf8) ?? ""

        if response.hasPrefix("bc_json(") {
            response = String(response.dropFirst("bc_json(".count).dropLast())
        }

        BCLog("Reponse from Blueconic: \(response)")

        var jsonError: Error?
        if let trimmedData: Data = response.data(using: .utf8), trimmedData.count > 0 {
            let jsonObject: Any?
            do {
                jsonObject = try JSONSerialization.jsonObject(with: trimmedData, options: JSONSerialization.ReadingOptions.mutableContainers)
            } catch {
                jsonError = error
                jsonObject = nil
            }

            // Check whether parsing was successful
            if jsonError != nil || jsonObject == nil {
                return (nil, jsonError)
            }

            // Expect an Array of method results
            if let jsonArray = jsonObject as? [Any] {
                for resultObject in jsonArray {
                    if let methodResult = resultObject as? [String: Any] {
                        if let methodId = methodResult[Constants.ID] as? String {
                            result[methodId] = methodResult
                        }
                    }
                }
                // Return the result for each call in a dictionary
                return (result, nil)
            } else if let result = jsonObject as? [String: Any] {
                return (result, nil)
            }
        }
        // Unable to parse the data
        return (nil, nil)
    }

    /**
     Creates the corresponding request object for the specified arguments.

     - parameter urlString: the base url
     - parameter commands: The array of method call objects that should be included in the request
     - parameter screenName: The screen name of the current view controller
     - returns: The request object for the specified parameters
    */
    func createRequest(_ urlString: String, commands: [BlueConicRequestCommand], screenName: String?) -> URLRequest? {
        // Start building the request object
        guard let url = URL(string: urlString) else {
            return nil
        }

        var request = URLRequest(url: url)

        // Set the user agent to this framework
        request.setValue(Constants.BLUECONIC_NAME, forHTTPHeaderField: Constants.Connector.USER_AGENT_FIELD)

        // Set the referrer data
        let referrerScreenName = screenName ?? ""
        let referrer: String = "app://\(BlueConicConfiguration.getPackageName())/\(referrerScreenName)"

        // Add the referrer to the HTTPHeader
        request.setValue(referrer, forHTTPHeaderField: Constants.Connector.REFERRER)
        request.setValue("close", forHTTPHeaderField: "Connection")
        // Set the timeout for the request to 30 seconds
        request.timeoutInterval = 30

        // Disable pipelingin
        request.httpShouldUsePipelining = false

        // Don't allow caching
        request.cachePolicy = .reloadIgnoringLocalCacheData

        // Use POST to avoid urls size limit issues
        request.httpMethod = METHOD_TYPE.POST.rawValue
        request.setValue(Constants.Connector.CONTENT_TYPE_VALUE, forHTTPHeaderField: Constants.Connector.CONTENT_TYPE_FIELD)

        if !BlueConicConfiguration.getShared() {
        
            // Add cookies manually
            let cookieStorage = self._session.configuration.httpCookieStorage as! BlueConicCookieStorage
            HTTPCookie.requestHeaderFields(with: cookieStorage.getCookies()).forEach { h in
                request.addValue(h.value, forHTTPHeaderField: h.key)
            }
        }
        
        if let headers = request.allHTTPHeaderFields {
            BCLog("Request headers to BlueConic: \(headers)")
        }
        request.httpBody = self.getPostData(commands)
        
        return request
    }

    /**
     Add a comment to this line
     Get the post query parameters

     - returns: The query parameters for the request url.
     */
    func getPostParameters() -> String {
        var parameters = "overruleReferrer=\(BlueConicConfiguration.getPackageName())"
        parameters = addSimulatorData(parameters)
        parameters = addTime(parameters)

        return parameters
    }

    /**
    Creates the corresponding request body for the specified arguments.

    - parameter commands: The array of method call objects that should be included in the request body
    - returns: The request body for the specified method calls
    */
    func getPostData(_ commands: [BlueConicRequestCommand]) -> Data {
        // create a list of json commands.

        var commandList: [String] = []
        for command in commands {
          commandList.append(command.toJson())
        }
        // add all json commands to the body seprated by a comma.
        let body: String = "[" + commandList.joined(separator: ",") + "]"

        BCLog("Request body to BlueConic: \(body)")
        return body.data(using: String.Encoding.utf8) ?? Data()
    }

    /**
     Create a set update command
     - parameter addProperties: A map of commit entries with the event type ADD.
     - parameter setProperties: A map of commit entries with the event type SET.
     - parameter events: A list of event commit entries.
     - returns: A list of request commands
     */
    class func setUpdateCommand(_ addProperties: [String: BlueConicCommitEntry], setProperties: [String: BlueConicCommitEntry], events: [BlueConicCommitEntry]) -> [BlueConicRequestCommand] {
        var commands = [BlueConicRequestCommand]()

        // Always send a getProfile call
        commands.append(BlueConicConnector.getProfileCommand(CallId.profile.rawValue))

        // Add a call if profile properties should be added
        if addProperties.count > 0 {
            commands.append(BlueConicConnector.getAddPropertiesCommand(CallId.addProperties.rawValue, properties: addProperties))
        }

        // Add a call if profile properties should be set
        if setProperties.count > 0 {
            commands.append(BlueConicConnector.getSetPropertiesCommand(CallId.setProperties.rawValue, properties: setProperties))
        }

        // Add a call if events should be created
        if events.count > 0 {
            for event in events {
                if let entry = event as? BlueConicEventCommitEntry {
                    let eventCount: Int = entry.getCount()
                    for index in 0 ..< eventCount {
                        commands.append(BlueConicConnector.getEventCommand(id: index + (CallId.events.rawValue), eventType: entry.getType(), interactionId: entry.getId()))
                    }
                }
            }
        }

        return commands
    }

    /**
     Creates a BlueConicRequestCommand, which can be used by the connector to send and retrieve data.
     - parameter id: Int is an unique id that can be used when retrieving data.
     - returns: a BlueConicRequestCommand that contains the profileParameters.
    */
    class func getProfileCommand(_ id: Int) -> BlueConicRequestCommand {
        // Create a parameter list with PAGEVIEW as object
        let parameterList = [Constants.ProfileProperties.PARAMETER_VALUE]

        // Add the profile parameters with forceCreate as key
        var profileParameters = [String: [String]]()
        profileParameters[Constants.ProfileProperties.PARAMETER_KEY] = parameterList

        // Create a method call with get profile as request
        return BlueConicRequestCommand(methodName: Constants.Calls.GET_PROFILE, callId: id, parameters: profileParameters, complexParameters: false)
    }

    /**
     Return the profile properties.
     - parameter id: The id for the request command
     - parameter hash: optional hash string of the values currently in the cache. When present, only profile properties with a different value are returned.
     - returns: The map with profile properties.
    */
    class func getGetPropertiesCommand(_ id: Int, hash: String?) -> BlueConicRequestCommand {
        // Add the profile parameters with forceCreate as key
        var parameters = [String: [String]]()
        if hash != nil {
            parameters["hash"] = [hash!]
        }
        // Create a method call with getProperties as request
        return BlueConicRequestCommand(methodName: Constants.Calls.GET_PROPERTIES, callId: id, parameters: parameters as [String: Any]?, complexParameters: false)
    }

    /**
     Getter for the add properties command.
     - parameter id: The unique command id.
     - parameter properties: A list of commit entries with the event type ADD.
     - returns: A request command.
     */
    class func getAddPropertiesCommand(_ id: Int, properties: [String: BlueConicCommitEntry]) -> BlueConicRequestCommand {
        return BlueConicRequestCommand(methodName: Constants.Calls.ADD_PROPERTIES, callId: id, parameters: properties, complexParameters: true)
    }

    /**
     Getter for the set properties command.
     - parameter id: The unique command id.
     - parameter properties: A list of commit entries with the event type SET.
     - returns: A request command.
     */
    class func getSetPropertiesCommand(_ id: Int, properties: [String: BlueConicCommitEntry]) -> BlueConicRequestCommand {
        return BlueConicRequestCommand(methodName: Constants.Calls.SET_PROPERTIES, callId: id, parameters: properties, complexParameters: true)
    }
    
    /**
     Creates a BlueConicRequestCommand, which can be used by the connector to send and retrieve data.
     - parameter id: Int is an unique id that can be used when retrieving data.
     - parameter eventType: The type of the event, PAGEVIEW
     - returns: a BlueConicRequestCommand that contains a create event that retrieve the interactions.
    */
    class func getInteractionCommand(_ id: Int, eventType: String) -> BlueConicRequestCommand {
        return BlueConicConnector.getInteractionCommand(id, eventType: eventType, properties: nil)
    }

    /**
     Creates a BlueConicRequestCommand, which can be used by the connector to send and retrieve data.
     - parameter id: Int is an unique id that can be used when retrieving data.
     - parameter eventType: The type of the event, PAGEVIEW
     - parameter eventType: The context for timeline events
     - returns: a BlueConicRequestCommand that contains a create event that retrieve the interactions.
    */
    class func getInteractionCommand(_ id: Int, eventType: String, properties: [String: Any]?) -> BlueConicRequestCommand {

        // Add the PAGEVIEW with forceCreate as key and add an empty string array with intraction as key
        var interactionParameters = [String: [String]]()
        interactionParameters[Constants.Interactions.PARAMETER_KEY] = [eventType]
        interactionParameters[Constants.Interactions.INTERACTION_KEY] = []
        
        if let properties = properties,
            let context = BlueConicConnector.convertDataToJSON(properties) {
            interactionParameters[Constants.Interactions.TIMELINE_CONTEXT_KEY] = [context]
            
        }

        // Create a method call with create event as request
        return BlueConicRequestCommand(methodName: Constants.Calls.CREATE_EVENT, callId: id, parameters: interactionParameters, complexParameters: false)
    }

    /*
     Creates a BlueConicRequestCommand, which can be used by the connector to send and retrieve data.
     - parameter id: Is an unique id that can be used when retrieving data.
     - parameter eventType: The type of the event added to the parameterlist {VIEW, CLICK, CONVERSION}
     - parameter interactionId: The interaction identifier
     - returns: a BlueConicRequestCommand that contains a createEvent.
    */
    class func getEventCommand(id: Int, eventType: String, interactionId: String) -> BlueConicRequestCommand {
        var eventParameters = [String: [String]]()
        eventParameters[Constants.Interactions.PARAMETER_KEY] = [eventType]
        eventParameters[Constants.Interactions.INTERACTION_KEY] = [interactionId]

        return BlueConicRequestCommand(methodName: Constants.Calls.CREATE_EVENT, callId: id, parameters: eventParameters, complexParameters: false)
    }

    /**
     Adds username and mobileSessionId information to the parameters, so they can be sent in the
     request to connect to the simulator.
     - parameter body: The request body.
     - returns: The body with added simulator data.
    */
    func addSimulatorData(_ body: String) -> String {
        if let username = self._userName,
            let mobileSessionId = self._mobileSessionId,
            !username.isEmpty,
            !mobileSessionId.isEmpty {
                return body + "&username=\(username)&mobileSessionId=\(mobileSessionId)"
        }
        return body
    }

    /**
     Adds username and mobileSessionId information to the parameters, so they can be sent in the
     request to connect to the simulator.
     - parameter body: The request body.
     - returns: The body with current time.
    */
    private func addTime(_ body: String) -> String {
        let date = Date()
        let calendar = Calendar.current
        let components = calendar.dateComponents([.hour, .minute], from: date)
        if let hour = components.hour,
           let minute = components.minute {
            return body + "&time=\(hour):\(minute)"
        }

        return body
    }

    private func BCLog(_ message: String, flag: String = Constants.Flag.INFO) {
        BlueConicConfiguration.log(message, tag: Constants.Debug.CONNECTOR, flag: flag)
    }
    
    /**
    Convert a dictionary to a JSON String
    - parameter values: A dictionary containing values.
    */
    class func convertDataToJSON(_ values: [String: Any]?) -> String? {
        // Convert the paramters values to JSON Data
        if let dictionary = values {
            var dictionaryData: Data?
            do {
                dictionaryData = try JSONSerialization.data(withJSONObject: dictionary, options: [])
            } catch {
                dictionaryData = nil
            }
            // Check if the data is correct and encode it to UTF-8
            if let data = dictionaryData, let result = String(data: data, encoding: .utf8) {
                // Return the JSON String
                return  result
            }
        }

        // Return nil, when the dictionary is unable to convert to JSON
        return nil
    }
}
