/*
* $LastChangedBy$
* $LastChangedDate$
* $LastChangedRevision$
* $HeadURL$
*
* Copyright 2014 BlueConic Inc./BlueConic B.V. All rights reserved.
*/

import Foundation

class BlueConicRequestCommand {

    // Private variables
    private var _methodName: String!
    private var _callId: String!
    private var _parameters: [String: Any]?
    private var _complexParameters: Bool = false

    /**
    Constructs a method call with the specified method name.

    - parameter methodName: The method name of the call, eg getProperties
    - parameter callId: The identifier for this method call
    */
    convenience init(methodName: String, callId: Int) {
        self.init(methodName: methodName, callId: callId, parameters: nil, complexParameters: false)
    }

    /**
    Constructs a method call with the specified method name and parameters.

    - parameter methodName: The method name of the call, eg "getProperties"
    - parameter callId: The identifier for this method call
    - parameter parameters: The map of parameters names to corresponding values, possibly nil.
    - parameter complexParameters: Indicates whether the parameters are complex parameters.
    */
    init(methodName: String, callId: Int, parameters: [String: Any]?, complexParameters: Bool) {
        self._methodName = methodName
        self._callId = "\(callId)" as String
        self._parameters = parameters
        self._complexParameters = complexParameters
    }

    /**
    Creates the json String representation for this method call.

    - returns:  the json String representation for this method call.
    */
    func toJson() -> String {
        // Create a new dictionary to store the method call attributes
        var dictionary = [String: Any]()
        dictionary[Constants.METHOD] = self._methodName
        dictionary[Constants.Calls.JSON_KEY] = Constants.Calls.JSON_VALUE
        dictionary[Constants.ID] = self._callId

        if let parameters = self._parameters {
            var parametersValues: [String: Any]
            // Add parameters to the dictionary
            if self._complexParameters {

                parametersValues = [Constants.PROPERTIES: self.getValuesOfEntries(parameters)]
            } else {
                parametersValues = parameters
            }

            // Convert the parameters to JSON
            if let parametersString = self.convertDataToJSON(values: parametersValues) {
                dictionary[Constants.PARAMS] =  parametersString
            }
        }

        // Convert the entire request to JSON
        if let dictionaryString = self.convertDataToJSON(values: dictionary) {
            return dictionaryString
        }

        // Return an empty string if it failed to convert it to correct JSON
        return ""
    }

    /**
    Convert a dictionary to a JSON String
    - parameter values: A dictionary containing values.
    */
    func convertDataToJSON(values: [String: Any]?) -> String? {
        // Convert the paramters values to JSON Data
        if let dictionary = values {
            var dictionaryData: Data?
            do {
                dictionaryData = try JSONSerialization.data(withJSONObject: dictionary, options: [])
            } catch {
                dictionaryData = nil
            }
            // Check if the data is correct and encode it to UTF-8
            if let data = dictionaryData, let result = String(data: data, encoding: .utf8) {
                // Return the JSON String
                return  result
            }
        }

        // Return nil, when the dictionary is unable to convert to JSON
        return nil
    }

    /**
    Returns a Dictionary with all values of each PropertyCommitEntry
    This function is needed to validate a complexParameters JSON Serialization.

    - parameter parameters: a Dictionary with BlueConicPropertyCommitEntries
    - returns: A Dictionary with the values as an Array of each BlueConicPropertyCommitEntry
    */
    func getValuesOfEntries(_ parameters: [String: Any]) -> [String: Any] {
        var result = [String: Any]()
        for (key, value) in  parameters {
            if let entry = value as? BlueConicPropertyCommitEntry {
                result[key] = entry.getValues()
            }
        }
        return result
    }
}
